package pneumaticCraft.api.heat;

import net.minecraft.block.state.IBlockState;
import net.minecraft.nbt.NBTTagCompound;
import net.minecraft.tileentity.TileEntity;
import net.minecraft.util.BlockPos;
import net.minecraft.world.World;

/**
 * Extend this class, and register it via {@link PneumaticRegistry.getInstance().registerHeatBehaviour()}
 * This can be used to add heat dependent logic to non-TE's or blocks you don't have access to. PneumaticCraft uses this to power Furnaces with heat,
 * and to turn Lava into Obsidian when heat is drained. This only works for ticking heat logic, so not for static heat sources like lava blocks.
 */
public abstract class HeatBehaviour<Tile extends TileEntity> {

    private IHeatExchangerLogic connectedHeatLogic;
    private World world;
    private BlockPos pos;
    private Tile cachedTE;
    private IBlockState block;

    /**
     * Called by the connected IHeatExchangerLogic.
     * @param connectedHeatLogic
     * @param world
     * @param pos
     */
    public void initialize(IHeatExchangerLogic connectedHeatLogic, World world, BlockPos pos){
        this.connectedHeatLogic = connectedHeatLogic;
        this.world = world;
        this.pos = pos;
        cachedTE = null;
        block = null;
    }

    public IHeatExchangerLogic getHeatExchanger(){
        return connectedHeatLogic;
    }

    public World getWorld(){
        return world;
    }

    public BlockPos getPos(){
        return pos;
    }

    public Tile getTileEntity(){
        if(cachedTE == null || cachedTE.func_145837_r()) cachedTE = (Tile)world.func_175625_s(pos);
        return cachedTE;
    }

    public IBlockState getBlockState(){
        if(block == null) block = world.func_180495_p(pos);
        return block;
    }

    /**
     * Unique id for this behaviour. Used in NBT saving. I recommend prefixing it with your modid.
     * @return
     */
    public abstract String getId();

    /**
     * Return true when this heat behaviour is applicable for this coordinate. World access methods can be used here (getWorld(), getX(), getY(), getZ(), getBlock(), getTileEntity()).
     * @return
     */
    public abstract boolean isApplicable();

    /**
     * Called every tick to update this behaviour.
     */
    public abstract void update();

    public void writeToNBT(NBTTagCompound tag){
        tag.func_74768_a("x", pos.func_177958_n());
        tag.func_74768_a("y", pos.func_177956_o());
        tag.func_74768_a("z", pos.func_177952_p());
    }

    public void readFromNBT(NBTTagCompound tag){
        pos = new BlockPos(tag.func_74762_e("x"), tag.func_74762_e("y"), tag.func_74762_e("z"));
    }

    @Override
    public boolean equals(Object o){
        if(o instanceof HeatBehaviour) {
            HeatBehaviour behaviour = (HeatBehaviour)o;
            return behaviour.getId().equals(getId()) && behaviour.getPos().equals(getPos());
        } else {
            return false;
        }
    }

    @Override
    public int hashCode(){
        int i = getId().hashCode();
        i = i * 31 + getPos().hashCode();
        return i;
    }
}
