package pneumaticCraft.common.heat;

import java.util.ArrayList;
import java.util.HashSet;
import java.util.Iterator;
import java.util.List;
import java.util.Set;

import net.minecraft.nbt.NBTTagCompound;
import net.minecraft.nbt.NBTTagList;
import net.minecraft.util.BlockPos;
import net.minecraft.util.EnumFacing;
import net.minecraft.world.World;
import pneumaticCraft.api.heat.HeatBehaviour;
import pneumaticCraft.api.heat.IHeatExchangerLogic;
import pneumaticCraft.common.heat.behaviour.HeatBehaviourManager;
import pneumaticCraft.common.network.GuiSynced;

public class HeatExchangerLogic implements IHeatExchangerLogic{
    private final Set<IHeatExchangerLogic> hullExchangers = new HashSet<IHeatExchangerLogic>();
    private final Set<IHeatExchangerLogic> connectedExchangers = new HashSet<IHeatExchangerLogic>();
    private List<HeatBehaviour> behaviours = new ArrayList<HeatBehaviour>();
    private List<HeatBehaviour> newBehaviours;//Required to prevent a CME
    @GuiSynced
    private double temperature = 295;//degrees Kelvin, 20 degrees by default.
    private double thermalResistance = 1;
    private double thermalCapacity = 1;
    private static boolean isAddingOrRemovingLogic;

    @Override
    public void initializeAsHull(World world, BlockPos pos, EnumFacing... validSides){
        if(world.field_72995_K) return;
        for(IHeatExchangerLogic logic : hullExchangers) {
            removeConnectedExchanger(logic);
        }
        hullExchangers.clear();
        newBehaviours = new ArrayList<HeatBehaviour>();
        for(EnumFacing d : EnumFacing.field_82609_l) {
            if(isSideValid(validSides, d)) {
                HeatBehaviourManager.getInstance().addHeatBehaviours(world, pos.func_177972_a(d), this, newBehaviours);
                IHeatExchangerLogic logic = HeatExchangerManager.getInstance().getLogic(world, pos.func_177972_a(d), d.func_176734_d());
                if(logic != null) {
                    hullExchangers.add(logic);
                    addConnectedExchanger(logic);
                }
            }
        }
    }

    private boolean isSideValid(EnumFacing[] validSides, EnumFacing side){
        if(validSides.length == 0) return true;
        for(EnumFacing d : validSides) {
            if(d == side) return true;
        }
        return false;
    }

    @Override
    public void addConnectedExchanger(IHeatExchangerLogic exchanger){
        connectedExchangers.add(exchanger);
        if(!isAddingOrRemovingLogic) {
            isAddingOrRemovingLogic = true;
            exchanger.addConnectedExchanger(this);
            isAddingOrRemovingLogic = false;
        }
    }

    @Override
    public void removeConnectedExchanger(IHeatExchangerLogic exchanger){
        connectedExchangers.remove(exchanger);
        if(!isAddingOrRemovingLogic) {
            isAddingOrRemovingLogic = true;
            exchanger.removeConnectedExchanger(this);
            isAddingOrRemovingLogic = false;
        }
    }

    @Override
    public double getTemperature(){
        return temperature;
    }

    @Override
    public void setTemperature(double temperature){
        this.temperature = temperature;
    }

    @Override
    public void setThermalResistance(double thermalResistance){
        this.thermalResistance = thermalResistance;
    }

    @Override
    public double getThermalResistance(){
        return thermalResistance;
    }

    @Override
    public void setThermalCapacity(double capacity){
        thermalCapacity = capacity;
    }

    @Override
    public double getThermalCapacity(){
        return thermalCapacity;
    }

    @Override
    public void writeToNBT(NBTTagCompound tag){
        tag.func_74780_a("temperature", temperature);
        NBTTagList tagList = new NBTTagList();
        for(HeatBehaviour behaviour : behaviours) {
            NBTTagCompound t = new NBTTagCompound();
            t.func_74778_a("id", behaviour.getId());
            behaviour.writeToNBT(t);
            tagList.func_74742_a(t);
        }
        tag.func_74782_a("behaviours", tagList);
    }

    @Override
    public void readFromNBT(NBTTagCompound tag){
        temperature = tag.func_74769_h("temperature");
        behaviours.clear();
        NBTTagList tagList = tag.func_150295_c("behaviours", 10);
        for(int i = 0; i < tagList.func_74745_c(); i++) {
            NBTTagCompound t = tagList.func_150305_b(i);
            HeatBehaviour behaviour = HeatBehaviourManager.getInstance().getBehaviourForId(t.func_74779_i("id"));
            if(behaviour != null) {
                behaviour.readFromNBT(t);
                behaviours.add(behaviour);
            }
        }
    }

    @Override
    public void update(){
        if(getThermalCapacity() < 0.1D) {
            temperature = 295;
            return;
        }
        if(newBehaviours != null) {
            List<HeatBehaviour> oldBehaviours = behaviours;
            behaviours = newBehaviours;
            newBehaviours = null;
            for(HeatBehaviour oldBehaviour : oldBehaviours) {//Transfer over equal heat behaviour's info.
                int equalBehaviourIndex = behaviours.indexOf(oldBehaviour);
                if(equalBehaviourIndex >= 0) {
                    NBTTagCompound tag = new NBTTagCompound();
                    oldBehaviour.writeToNBT(tag);
                    behaviours.get(equalBehaviourIndex).readFromNBT(tag);
                }
            }
        }
        Iterator<HeatBehaviour> iterator = behaviours.iterator();
        while(iterator.hasNext()) {
            HeatBehaviour behaviour = iterator.next();
            if(behaviour.getWorld() != null) {//upon loading from NBT the world is null. gets initialized once 'initializeAsHull' is invoked.
                if(behaviour.isApplicable()) {
                    behaviour.update();
                } else {
                    iterator.remove();
                }
            }
        }
        for(IHeatExchangerLogic logic : connectedExchangers) {
            exchange(logic, this, getTickingHeatExchangers());//As the connected logics also will tick, we should prevent dispersing more when more are connected.
        }
    }

    public static void exchange(IHeatExchangerLogic logic, IHeatExchangerLogic logic2){
        exchange(logic, logic2, 1);
    }

    public static void exchange(IHeatExchangerLogic logic, IHeatExchangerLogic logic2, double dispersionDivider){
        if(logic.getThermalCapacity() < 0.1D) {
            logic.setTemperature(295);
            return;
        }
        double deltaTemp = logic.getTemperature() - logic2.getTemperature();

        double totalResistance = logic2.getThermalResistance() + logic.getThermalResistance();
        deltaTemp /= dispersionDivider;
        deltaTemp /= totalResistance;

        double maxDeltaTemp = (logic.getTemperature() * logic.getThermalCapacity() - logic2.getTemperature() * logic2.getThermalCapacity()) / 2;//Calculate the heat needed to exactly equalize the heat.
        if(maxDeltaTemp >= 0 && deltaTemp > maxDeltaTemp || maxDeltaTemp <= 0 && deltaTemp < maxDeltaTemp) deltaTemp = maxDeltaTemp;
        logic2.addHeat(deltaTemp);
        logic.addHeat(-deltaTemp);
    }

    private int getTickingHeatExchangers(){
        int tickingHeatExchangers = 1;
        for(IHeatExchangerLogic logic : connectedExchangers) {
            if(logic instanceof HeatExchangerLogic) tickingHeatExchangers++;
        }
        return tickingHeatExchangers;
    }

    @Override
    public void addHeat(double amount){
        temperature += amount / getThermalCapacity();
        temperature = Math.max(0, Math.min(2273, temperature));
    }

}
